@@ -8,8 +8,9 @@ Created on Tue Nov  3 15:07:16 2017
 from __future__ import division, print_function, unicode_literals, absolute_import
 import unittest
 import sys
-import numpy as np
 import os
+import numpy as np
+import dask.array as da
 import h5py
 sys.path.append("../../pyUSID/")
 from pyUSID.io import dtype_utils
@@ -58,7 +59,10 @@ class TestDtypeUtils(unittest.TestCase):
     def tearDown(self):
         os.remove(file_path)
 
-    def test_contains_integers(self):
+
+class TestContainsIntegers(unittest.TestCase):
+
+    def test_typical(self):
         self.assertTrue(dtype_utils.contains_integers([1, 2, -3, 4]))
         self.assertTrue(dtype_utils.contains_integers(range(5)))
         self.assertTrue(dtype_utils.contains_integers([2, 5, 8, 3], min_val=2))
@@ -74,7 +78,20 @@ class TestDtypeUtils(unittest.TestCase):
         with self.assertRaises(TypeError):
             _ = dtype_utils.contains_integers(14)
 
-    def test_integers_to_slices_illegal(self):
+    def test_illegal_min_val(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.contains_integers([1, 2, 3, 4], min_val='hello')
+
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.contains_integers([1, 2, 3, 4], min_val=[1, 2])
+
+        with self.assertRaises(ValueError):
+            _ = dtype_utils.contains_integers([1, 2, 3, 4], min_val=1.234)
+
+
+class TestIntegersToSlices(unittest.TestCase):
+
+    def test_illegal_inputs(self):
         with self.assertRaises(TypeError):
             dtype_utils.integers_to_slices(slice(1, 15))
         with self.assertRaises(ValueError):
@@ -82,83 +99,88 @@ class TestDtypeUtils(unittest.TestCase):
         with self.assertRaises(ValueError):
             dtype_utils.integers_to_slices(['asdds', None, True, 45.344, 5 + 6j])
 
-    def test_integers_to_slices_positive(self):
+    def test_positive(self):
         expected = [slice(0, 3), slice(7, 8), slice(14, 18), slice(22, 23), slice(27, 28), slice(29, 30), slice(31, 32)]
         inputs = np.hstack([range(item.start, item.stop) for item in expected])
         ret_val = dtype_utils.integers_to_slices(inputs)
         self.assertEqual(expected, ret_val)
 
-    def test_integers_to_slices_negative(self):
+    def test_negative(self):
         expected = [slice(-7, -4), slice(-2, 3), slice(14, 18), slice(22, 23), slice(27, 28), slice(29, 30)]
         inputs = np.hstack([range(item.start, item.stop) for item in expected])
         ret_val = dtype_utils.integers_to_slices(inputs)
         self.assertEqual(expected, ret_val)
 
-    def test_stack_real_to_complex_single(self):
+
+class TestStackRealToComplex(unittest.TestCase):
+
+    def test_single(self):
         expected = 4.32 + 5.67j
         real_val = [np.real(expected), np.imag(expected)]
         actual = dtype_utils.stack_real_to_complex(real_val)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_complex_to_real_1d_array(self):
-        complex_array = 5 * np.random.rand(5) + 7j * np.random.rand(5)
-        actual = dtype_utils.flatten_complex_to_real(complex_array)
-        expected = np.hstack([np.real(complex_array), np.imag(complex_array)])
-        self.assertTrue(np.allclose(actual, expected))
-
-    def test_stack_real_1d_to_complex_array(self):
+    def test_1d(self):
         expected = 5 * np.random.rand(6) + 7j * np.random.rand(6)
         real_val = np.hstack([np.real(expected), np.imag(expected)])
         actual = dtype_utils.stack_real_to_complex(real_val)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_complex_to_real_2d_array(self):
-        complex_array = 5 * np.random.rand(2, 3) + 7j * np.random.rand(2, 3)
-        actual = dtype_utils.flatten_complex_to_real(complex_array)
-        expected = np.hstack([np.real(complex_array), np.imag(complex_array)])
-        self.assertTrue(np.allclose(actual, expected))
-
-    def test_stack_real_2d_to_complex_array(self):
+    def test_2d(self):
         expected = 5 * np.random.rand(2, 8) + 7j * np.random.rand(2, 8)
         real_val = np.hstack([np.real(expected), np.imag(expected)])
         actual = dtype_utils.stack_real_to_complex(real_val)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_complex_to_real_nd_array(self):
-        complex_array = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
-        actual = dtype_utils.flatten_complex_to_real(complex_array)
-        expected = np.concatenate([np.real(complex_array), np.imag(complex_array)], axis=3)
-        self.assertTrue(np.allclose(actual, expected))
-
-    def test_complex_to_real_h5_legal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            h5_comp = h5_f['complex']
-            actual = dtype_utils.flatten_complex_to_real(h5_comp)
-            expected = np.concatenate([np.real(h5_comp[()]), np.imag(h5_comp[()])], axis=len(h5_comp.shape) - 1)
-            self.assertTrue(np.allclose(actual, expected))
-
-    def test_complex_to_real_h5_illegal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.flatten_complex_to_real(h5_f['real'])
-
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.flatten_complex_to_real(h5_f['compound'])
-
-    def test_stack_real_nd_to_complex_array(self):
+    def base_nd(self, lazy_in, lazy):
         expected = 5 * np.random.rand(2, 3, 5, 8) + 7j * np.random.rand(2, 3, 5, 8)
         real_val = np.concatenate([np.real(expected), np.imag(expected)], axis=3)
-        actual = dtype_utils.stack_real_to_complex(real_val)
+        if lazy_in:
+            real_val = da.from_array(real_val, chunks=real_val.shape)
+        actual = dtype_utils.stack_real_to_complex(real_val, lazy=lazy)
+        if lazy:
+            self.assertIsInstance(actual, da.core.Array)
+            actual = actual.compute()
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_stack_real_nd_to_complex_h5_legal(self):
+    def test_nd_numpy_not_lazy(self):
+        self.base_nd(False, False)
+
+    def test_nd_numpy_lazy(self):
+        self.base_nd(False, True)
+
+    def test_nd_dask(self):
+        self.base_nd(True, True)
+
+    def test_invalid_inputs(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.stack_real_to_complex("Hello")
+
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.stack_real_to_complex([1, {'a': 1}, 'a', True])
+
+
+class TestStackRealToComplexHDF5(TestDtypeUtils):
+
+    def base_h5_legal(self, lazy):
         with h5py.File(file_path, mode='r') as h5_f:
             h5_real = h5_f['real2']
             expected = h5_real[:, :, :3] + 1j * h5_real[:, :, 3:]
-            actual = dtype_utils.stack_real_to_complex(h5_real)
+            actual = dtype_utils.stack_real_to_complex(h5_real, lazy=lazy)
+            if lazy:
+                self.assertIsInstance(actual, da.core.Array)
+                actual = actual.compute()
+            else:
+                self.assertIsInstance(actual, np.ndarray)
             self.assertTrue(np.allclose(actual, expected))
 
-    def test_stack_real_nd_to_complex_h5_illegal(self):
+    def test_h5_as_numpy(self):
+        self.base_h5_legal(False)
+
+    def test_h5_as_dask(self):
+        self.base_h5_legal(True)
+
+    def test_h5_illegal(self):
         with h5py.File(file_path, mode='r') as h5_f:
             with self.assertRaises(TypeError):
                 _ = dtype_utils.stack_real_to_complex(h5_f['complex'])
@@ -166,7 +188,47 @@ class TestDtypeUtils(unittest.TestCase):
             with self.assertRaises(TypeError):
                 _ = dtype_utils.stack_real_to_complex(h5_f['compound'])
 
-    def test_complex_to_real_compound_illegal(self):
+    def test_illegal_odd_last_dim(self):
+        expected = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
+        real_val = np.concatenate([np.real(expected), np.imag(expected)[..., :-1]], axis=3)
+        with self.assertRaises(ValueError):
+            _ = dtype_utils.stack_real_to_complex(real_val)
+
+
+class TestFlattenComplexToReal(unittest.TestCase):
+
+    def test_1d_array(self):
+        complex_array = 5 * np.random.rand(5) + 7j * np.random.rand(5)
+        actual = dtype_utils.flatten_complex_to_real(complex_array)
+        expected = np.hstack([np.real(complex_array), np.imag(complex_array)])
+        self.assertTrue(np.allclose(actual, expected))
+
+    def test_2d_array(self):
+        complex_array = 5 * np.random.rand(2, 3) + 7j * np.random.rand(2, 3)
+        actual = dtype_utils.flatten_complex_to_real(complex_array)
+        expected = np.hstack([np.real(complex_array), np.imag(complex_array)])
+        self.assertTrue(np.allclose(actual, expected))
+
+    def test_nd_array_numpy(self):
+        complex_array = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
+        actual = dtype_utils.flatten_complex_to_real(complex_array)
+        expected = np.concatenate([np.real(complex_array), np.imag(complex_array)], axis=3)
+        self.assertTrue(np.allclose(actual, expected))
+
+    def test_nd_array_dask(self):
+        complex_array = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
+        da_comp = da.from_array(complex_array, chunks=complex_array.shape)
+        actual = dtype_utils.flatten_complex_to_real(da_comp)
+        self.assertIsInstance(actual, da.core.Array)
+        actual = actual.compute()
+        expected = np.concatenate([np.real(complex_array), np.imag(complex_array)], axis=3)
+        self.assertTrue(np.allclose(actual, expected))
+
+    def test_real_in(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.flatten_complex_to_real(np.arange(4))
+
+    def test_compound_illegal(self):
         num_elems = 5
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = np.random.random(size=num_elems)
@@ -175,17 +237,46 @@ class TestDtypeUtils(unittest.TestCase):
         with self.assertRaises(TypeError):
             _ = dtype_utils.flatten_complex_to_real(structured_array)
 
-    def test_stack_real_to_complex_illegal_odd_last_dim(self):
-        expected = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
-        real_val = np.concatenate([np.real(expected), np.imag(expected)[..., :-1]], axis=3)
-        with self.assertRaises(ValueError):
-            _ = dtype_utils.stack_real_to_complex(real_val)
 
-    def test_get_compund_sub_dtypes_legal(self):
+class TestFlattenComplexToRealHDF5(TestDtypeUtils):
+
+    def base_h5_legal(self, lazy):
+        with h5py.File(file_path, mode='r') as h5_f:
+            h5_comp = h5_f['complex']
+            actual = dtype_utils.flatten_complex_to_real(h5_comp, lazy=lazy)
+            if lazy:
+                self.assertIsInstance(actual, da.core.Array)
+                actual = actual.compute()
+            expected = np.concatenate([np.real(h5_comp[()]), np.imag(h5_comp[()])], axis=len(h5_comp.shape) - 1)
+            self.assertTrue(np.allclose(actual, expected))
+
+    def test_h5_legal_not_lazy(self):
+        self.base_h5_legal(False)
+
+    def test_h5_legal_lazy(self):
+        self.base_h5_legal(True)
+
+    def test_h5_illegal(self):
+        with h5py.File(file_path, mode='r') as h5_f:
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.flatten_complex_to_real(h5_f)
+
+    def test_h5_illegal_dtype(self):
+        with h5py.File(file_path, mode='r') as h5_f:
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.flatten_complex_to_real(h5_f['real'])
+
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.flatten_complex_to_real(h5_f['compound'])
+
+
+class TestGetCompoundSubTypes(unittest.TestCase):
+
+    def test_legal(self):
         self.assertEqual({'r': np.float32, 'g': np.uint16, 'b': np.float64},
                          dtype_utils.get_compound_sub_dtypes(struc_dtype))
 
-    def test_get_compound_sub_dtypes_illegal(self):
+    def test_illegal(self):
         with self.assertRaises(TypeError):
             _ = dtype_utils.get_compound_sub_dtypes(np.float16)
 
@@ -195,7 +286,82 @@ class TestDtypeUtils(unittest.TestCase):
         with self.assertRaises(TypeError):
             _ = dtype_utils.get_compound_sub_dtypes(np.arange(4))
 
-    def test_compound_to_real_single(self):
+
+class TestStackRealToCompound(unittest.TestCase):
+
+    def test_single(self):
+        num_elems = 1
+        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
+        structured_array['r'] = r_vals = np.random.random(size=num_elems)
+        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
+        structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        real_val = np.concatenate((r_vals, g_vals, b_vals))
+        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype)
+        self.assertTrue(compare_structured_arrays(actual, structured_array[0]))
+
+    def test_1d(self):
+        num_elems = 5
+        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
+        structured_array['r'] = r_vals = np.random.random(size=num_elems)
+        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
+        structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        real_val = np.concatenate((r_vals, g_vals, b_vals))
+        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype)
+        self.assertTrue(compare_structured_arrays(actual, structured_array))
+
+    def test_1d_list(self):
+        num_elems = 5
+        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
+        structured_array['r'] = r_vals = np.random.random(size=num_elems)
+        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
+        structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        real_val = np.concatenate((r_vals, g_vals, b_vals))
+        actual = dtype_utils.stack_real_to_compound(list(real_val), struc_dtype)
+        self.assertTrue(compare_structured_arrays(actual, structured_array))
+
+    def base_nd(self, in_lazy, out_lazy):
+        num_elems = (2, 3, 5, 7)
+        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
+        structured_array['r'] = r_vals = np.random.random(size=num_elems)
+        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
+        structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        real_val = np.concatenate((r_vals, g_vals, b_vals), axis=len(num_elems) - 1)
+        if in_lazy:
+            real_val = da.from_array(real_val, chunks=real_val.shape)
+        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype, lazy=out_lazy)
+        if out_lazy:
+            self.assertIsInstance(actual, da.core.Array)
+            actual = actual.compute()
+        self.assertTrue(compare_structured_arrays(actual, structured_array))
+
+    def test_nd_dask_in(self):
+        with self.assertRaises(NotImplementedError):
+            self.base_nd(True, True)
+
+    def test_nd_numpy_in_not_lazy(self):
+        self.base_nd(False, False)
+
+    def test_nd_numpy_in_lazy(self):
+        with self.assertRaises(NotImplementedError):
+            self.base_nd(False, True)
+
+    def test_illegal(self):
+        num_elems = (3, 5)
+        r_vals = np.random.random(size=num_elems)
+
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.stack_real_to_compound(r_vals, np.float32)
+
+        with self.assertRaises(ValueError):
+            _ = dtype_utils.stack_real_to_compound(r_vals, struc_dtype)
+
+        with self.assertRaises(ValueError):
+            _ = dtype_utils.stack_real_to_compound([1, 'a', {'a': 1}, False], struc_dtype)
+
+
+class TestFlattenCompoundToReal(unittest.TestCase):
+
+    def test_single(self):
         num_elems = 1
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -205,17 +371,7 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.flatten_compound_to_real(structured_array[0])
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_real_to_compound_single(self):
-        num_elems = 1
-        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
-        structured_array['r'] = r_vals = np.random.random(size=num_elems)
-        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
-        structured_array['b'] = b_vals = np.random.random(size=num_elems)
-        real_val = np.concatenate((r_vals, g_vals, b_vals))
-        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype)
-        self.assertTrue(compare_structured_arrays(actual, structured_array[0]))
-
-    def test_compound_to_real_1d(self):
+    def test_1d(self):
         num_elems = 5
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -225,72 +381,31 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.flatten_compound_to_real(structured_array)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_real_to_compound_1d(self):
-        num_elems = 5
-        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
-        structured_array['r'] = r_vals = np.random.random(size=num_elems)
-        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
-        structured_array['b'] = b_vals = np.random.random(size=num_elems)
-        real_val = np.concatenate((r_vals, g_vals, b_vals))
-        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype)
-        self.assertTrue(compare_structured_arrays(actual, structured_array))
-
-    def test_compound_to_real_nd(self):
+    def base_nd(self, lazy_in, lazy):
         num_elems = (5, 7, 2, 3)
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
         structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
         structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        if lazy_in:
+            structured_array = da.from_array(structured_array, chunks=structured_array.shape)
         expected = np.concatenate((r_vals, g_vals, b_vals), axis=len(num_elems) - 1)
-        actual = dtype_utils.flatten_compound_to_real(structured_array)
+        actual = dtype_utils.flatten_compound_to_real(structured_array, lazy=lazy)
+        if lazy:
+            self.assertIsInstance(actual, da.core.Array)
+            actual = actual.compute()
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_compound_to_real_nd_h5_legal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            h5_comp = h5_f['compound']
-            actual = dtype_utils.flatten_compound_to_real(h5_comp)
-            expected = np.concatenate([h5_comp['r'], h5_comp['g'], h5_comp['b']], axis=len(h5_comp.shape) - 1)
-            self.assertTrue(np.allclose(actual, expected))
+    def test_numpy_nd_not_lazy(self):
+        self.base_nd(False, False)
 
-    def test_compound_to_real_nd_h5_illegal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.flatten_compound_to_real(h5_f['real'])
+    def test_numpy_nd_lazy(self):
+        self.base_nd(False, True)
 
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.flatten_compound_to_real(h5_f['complex'])
+    def test_dask_nd(self):
+        self.base_nd(True, True)
 
-    def test_real_to_compound_nd(self):
-        num_elems = (2, 3, 5, 7)
-        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
-        structured_array['r'] = r_vals = np.random.random(size=num_elems)
-        structured_array['g'] = g_vals = np.random.randint(0, high=1024, size=num_elems)
-        structured_array['b'] = b_vals = np.random.random(size=num_elems)
-        real_val = np.concatenate((r_vals, g_vals, b_vals), axis=len(num_elems) - 1)
-        actual = dtype_utils.stack_real_to_compound(real_val, struc_dtype)
-        self.assertTrue(compare_structured_arrays(actual, structured_array))
-
-    def test_real_to_compound_nd_h5_legal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            h5_real = h5_f['real2']
-            structured_array = np.zeros(shape=list(h5_real.shape)[:-1] + [h5_real.shape[-1] // len(struc_dtype.names)],
-                                        dtype=struc_dtype)
-            for name_ind, name in enumerate(struc_dtype.names):
-                i_start = name_ind * structured_array.shape[-1]
-                i_end = (name_ind + 1) * structured_array.shape[-1]
-                structured_array[name] = h5_real[..., i_start:i_end]
-            actual = dtype_utils.stack_real_to_compound(h5_real, struc_dtype)
-            self.assertTrue(compare_structured_arrays(actual, structured_array))
-
-    def test_real_to_compound_nd_h5_illegal(self):
-        with h5py.File(file_path, mode='r') as h5_f:
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.stack_real_to_compound(h5_f['compound'], struc_dtype)
-
-            with self.assertRaises(TypeError):
-                _ = dtype_utils.stack_real_to_compound(h5_f['complex'], struc_dtype)
-
-    def test_compound_to_real_illegal(self):
+    def test_illegal(self):
         num_elems = (2, 3)
         r_vals = np.random.random(size=num_elems)
         with self.assertRaises(TypeError):
@@ -299,27 +414,87 @@ class TestDtypeUtils(unittest.TestCase):
         with self.assertRaises(TypeError):
             _ = dtype_utils.flatten_compound_to_real(14)
 
-    def test_real_to_compound_illegal(self):
-        num_elems = (3, 5)
-        r_vals = np.random.random(size=num_elems)
-        with self.assertRaises(TypeError):
-            _ = dtype_utils.stack_real_to_compound(r_vals, np.float32)
-        with self.assertRaises(ValueError):
-            _ = dtype_utils.stack_real_to_compound(r_vals, struc_dtype)
 
-    def test_flatten_to_real_complex_nd(self):
+class TestFlattenCompoundToRealHDF5(TestDtypeUtils):
+
+    def base_nd_h5_legal(self, lazy):
+        with h5py.File(file_path, mode='r') as h5_f:
+            h5_comp = h5_f['compound']
+            actual = dtype_utils.flatten_compound_to_real(h5_comp, lazy=lazy)
+            if lazy:
+                self.assertIsInstance(actual, da.core.Array)
+                actual = actual.compute()
+            expected = np.concatenate([h5_comp['r'], h5_comp['g'], h5_comp['b']], axis=len(h5_comp.shape) - 1)
+            self.assertTrue(np.allclose(actual, expected))
+
+    def test_h5_legal_not_lazy(self):
+        self.base_nd_h5_legal(False)
+
+    def test_h5_legal_lazy(self):
+        self.base_nd_h5_legal(True)
+
+    def test_nd_h5_illegal(self):
+        with h5py.File(file_path, mode='r') as h5_f:
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.flatten_compound_to_real(h5_f['real'])
+
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.flatten_compound_to_real(h5_f['complex'])
+
+
+class TestStackRealToCompoundHDF5(TestDtypeUtils):
+
+    def base_nd_h5_legal(self, lazy):
+        with h5py.File(file_path, mode='r') as h5_f:
+            h5_real = h5_f['real2']
+            structured_array = np.zeros(shape=list(h5_real.shape)[:-1] + [h5_real.shape[-1] // len(struc_dtype.names)],
+                                        dtype=struc_dtype)
+            for name_ind, name in enumerate(struc_dtype.names):
+                i_start = name_ind * structured_array.shape[-1]
+                i_end = (name_ind + 1) * structured_array.shape[-1]
+                structured_array[name] = h5_real[..., i_start:i_end]
+            actual = dtype_utils.stack_real_to_compound(h5_real, struc_dtype, lazy=lazy)
+            if lazy:
+                self.assertIsInstance(actual, da.core.Array)
+                actual = actual.compute()
+            self.assertTrue(compare_structured_arrays(actual, structured_array))
+
+    def test_nd_h5(self):
+        self.base_nd_h5_legal(False)
+
+    def test_nd_h5_lazy(self):
+        with self.assertRaises(NotImplementedError):
+            self.base_nd_h5_legal(True)
+
+    def test_nd_h5_illegal(self):
+        with h5py.File(file_path, mode='r') as h5_f:
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.stack_real_to_compound(h5_f['compound'], struc_dtype)
+
+            with self.assertRaises(TypeError):
+                _ = dtype_utils.stack_real_to_compound(h5_f['complex'], struc_dtype)
+
+
+class TestFlattenToRealNumpy(unittest.TestCase):
+
+    def test_real_nd(self):
+        real_array = 5 * np.random.rand(2, 3, 5, 7)
+        actual = dtype_utils.flatten_to_real(real_array)
+        self.assertTrue(np.allclose(actual, real_array))
+
+    def test_complex_nd(self):
         complex_array = 5 * np.random.rand(2, 3, 5, 7) + 7j * np.random.rand(2, 3, 5, 7)
         actual = dtype_utils.flatten_to_real(complex_array)
         expected = np.concatenate([np.real(complex_array), np.imag(complex_array)], axis=3)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_flatten_to_real_complex_single(self):
+    def test_complex_single(self):
         complex_val = 4.32 + 5.67j
         expected = [np.real(complex_val), np.imag(complex_val)]
         actual = dtype_utils.flatten_to_real(complex_val)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_flatten_to_real_compound_nd(self):
+    def test_compound_nd(self):
         num_elems = (5, 7, 2, 3)
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -329,7 +504,7 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.flatten_to_real(structured_array)
         self.assertTrue(np.allclose(actual, expected))
 
-    def test_flatten_to_real_compound_single(self):
+    def test_compound_single(self):
         num_elems = 1
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -339,19 +514,22 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.flatten_to_real(structured_array[0])
         self.assertTrue(np.allclose(actual, expected))
 
-    def stack_real_to_target_complex_single(self):
+
+class TestStackRealToTargetDtypeNumpy(unittest.TestCase):
+
+    def test_complex_single(self):
         expected = 4.32 + 5.67j
         real_val = [np.real(expected), np.imag(expected)]
         actual = dtype_utils.stack_real_to_target_dtype(real_val, np.complex)
         self.assertTrue(np.allclose(actual, expected))
 
-    def stack_real_to_target_complex_nd(self):
+    def test_complex_nd(self):
         expected = 5 * np.random.rand(2, 3, 5, 8) + 7j * np.random.rand(2, 3, 5, 8)
         real_val = np.concatenate([np.real(expected), np.imag(expected)], axis=3)
         actual = dtype_utils.stack_real_to_target_dtype(real_val, np.complex)
         self.assertTrue(np.allclose(actual, expected))
 
-    def stack_real_to_target_compound_single(self):
+    def test_compound_single(self):
         num_elems = 1
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -361,7 +539,7 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.stack_real_to_target_dtype(real_val, struc_dtype)
         self.assertTrue(compare_structured_arrays(actual, structured_array[0]))
 
-    def stack_real_to_target_compound_nd(self):
+    def test_compound_nd(self):
         num_elems = (2, 3, 5, 7)
         structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
         structured_array['r'] = r_vals = np.random.random(size=num_elems)
@@ -371,7 +549,40 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.stack_real_to_target_dtype(real_val, struc_dtype)
         self.assertTrue(compare_structured_arrays(actual, structured_array))
 
-    def test_check_dtype_real_numpy(self):
+
+    def test_compound_nd(self):
+        num_elems = (2, 3, 5, 7)
+        structured_array = np.zeros(shape=num_elems, dtype=struc_dtype)
+        structured_array['r'] = r_vals = np.random.random(size=num_elems)
+        structured_array['g'] = g_vals = np.random.randint(0, high=1024,
+                                                           size=num_elems)
+        structured_array['b'] = b_vals = np.random.random(size=num_elems)
+        real_val = np.concatenate((r_vals, g_vals, b_vals),
+                                  axis=len(num_elems) - 1)
+        actual = dtype_utils.stack_real_to_target_dtype(real_val, struc_dtype)
+        self.assertTrue(compare_structured_arrays(actual, structured_array))
+
+    def test_cast(self):
+        uint_array = np.random.randint(0, high=15, size=(3, 4, 5),
+                                       dtype=np.uint16)
+        actual = dtype_utils.stack_real_to_target_dtype(uint_array, np.float32)
+        self.assertTrue(np.allclose(actual, uint_array))
+
+
+class TestStackRealToTargetDtypeHDF5(TestDtypeUtils):
+
+    def test_stack_real_nd_to_target_complex_h5_legal(self):
+        with h5py.File(file_path, mode='r') as h5_f:
+            h5_real = h5_f['real2']
+            expected = h5_real[:, :, :3] + 1j * h5_real[:, :, 3:]
+            actual = dtype_utils.stack_real_to_target_dtype(h5_real,
+                                                            np.complex)
+            self.assertTrue(np.allclose(actual, expected))
+
+
+class TestCheckDtype(TestDtypeUtils):
+
+    def test_real_numpy(self):
         # real_matrix = np.random.rand(5, 7)
         # func, is_complex, is_compound, n_features, type_mult = dtype_utils.check_dtype(real_matrix)
         with h5py.File(file_path, mode='r') as h5_f:
@@ -382,7 +593,7 @@ class TestDtypeUtils(unittest.TestCase):
             self.assertEqual(n_features, h5_f['real'].shape[1])
             self.assertEqual(type_mult, h5_f['real'].dtype.type(0).itemsize)
 
-    def test_check_dtype_complex_numpy(self):
+    def test_complex_numpy(self):
         with h5py.File(file_path, mode='r') as h5_f:
             func, is_complex, is_compound, n_features, type_mult = dtype_utils.check_dtype(h5_f['complex'])
             self.assertEqual(func, dtype_utils.flatten_complex_to_real)
@@ -391,7 +602,7 @@ class TestDtypeUtils(unittest.TestCase):
             self.assertEqual(n_features, 2 * h5_f['complex'].shape[1])
             self.assertEqual(type_mult, 2 * np.real(h5_f['complex'][0, 0]).dtype.itemsize)
 
-    def test_check_dtype_compound_numpy(self):
+    def test_compound_numpy(self):
         with h5py.File(file_path, mode='r') as h5_f:
             func, is_complex, is_compound, n_features, type_mult = dtype_utils.check_dtype(h5_f['compound'])
             self.assertEqual(func, dtype_utils.flatten_compound_to_real)
@@ -400,7 +611,14 @@ class TestDtypeUtils(unittest.TestCase):
             self.assertEqual(n_features, 3 * h5_f['compound'].shape[1])
             self.assertEqual(type_mult, 3 * np.float32(0).itemsize)
 
-    def test_get_compound_sub_dtypes_valid(self):
+    def test_non_hdf(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.check_dtype(np.arange(15))
+
+
+class TestGetCompoundSubTypes(unittest.TestCase):
+
+    def test_valid(self):
         input_dict = {'names': ['r', 'g', 'b'], 'formats': [np.float32, np.uint16, np.float64]}
         expected = dict()
         for name, dtype in zip(input_dict['names'], input_dict['formats']):
@@ -409,23 +627,29 @@ class TestDtypeUtils(unittest.TestCase):
         actual = dtype_utils.get_compound_sub_dtypes(struct_dtype)
         self.assertEqual(expected, actual)
 
-    def test_get_compound_sub_dtypes_invalid(self):
+    def test_invalid(self):
         for item in [np.float16, np.complex, 4, 2.343, True, 'ssdsds', np.arange(5)]:
             with self.assertRaises(TypeError):
                 _ = dtype_utils.get_compound_sub_dtypes(item)
 
-    def test_validate_dtype_valid(self):
+
+class TestValidateDtype(unittest.TestCase):
+
+    def test_valid(self):
         struct_dtype = np.dtype({'names': ['r', 'g', 'b'],
                                 'formats': [np.float32, np.uint16, np.float64]})
         for dtype in [np.float32, np.float16, np.complex, np.complex64, np.uint8, np.int16, struct_dtype]:
             self.assertTrue(dtype_utils.validate_dtype(dtype))
 
-    def test_validate_dtype_invalid(self):
+    def test_invalid(self):
         for dtype in [6, 'dssds', np.arange(5), True]:
             with self.assertRaises(TypeError):
                 dtype_utils.validate_dtype(dtype)
 
-    def test_is_complex_dtype_valid(self):
+
+class TestIsComplexDtype(unittest.TestCase):
+
+    def test_valid(self):
         struct_dtype = np.dtype({'names': ['r', 'g', 'b'],
                                  'formats': [np.float32, np.uint16, np.float64]})
         for dtype in [np.float32, np.float16, np.uint8, np.int16, struct_dtype, bool]:
@@ -435,5 +659,76 @@ class TestDtypeUtils(unittest.TestCase):
             self.assertTrue(dtype_utils.is_complex_dtype(dtype))
 
 
+class TestGetExponent(unittest.TestCase):
+
+    def test_negative_small(self):
+        expected = -7
+        self.assertEqual(expected,
+                         dtype_utils.get_exponent(np.arange(5) * -10**expected))
+
+    def test_positive_large(self):
+        expected = 4
+        self.assertEqual(expected,
+                         dtype_utils.get_exponent(np.arange(6) * 10 ** expected))
+
+    def test_mixed_large(self):
+        expected = 4
+        self.assertEqual(expected,
+                         dtype_utils.get_exponent(np.random.randint(-8, high=3, size=(5, 5)) * 10 ** expected))
+
+    def test_illegal_type(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.get_exponent('hello')
+            _ = dtype_utils.get_exponent([1, 2, 3])
+
+
+class TestValidateStringArgs(unittest.TestCase):
+
+    def test_empty(self):
+        with self.assertRaises(ValueError):
+            _ = dtype_utils.validate_string_args(['      '], ['meh'])
+
+    def test_spaces(self):
+        expected = 'fd'
+        [ret] = dtype_utils.validate_string_args(['  ' + expected + '    '], ['meh'])
+        self.assertEqual(expected, ret)
+
+    def test_single(self):
+        expected = 'fd'
+        [ret] = dtype_utils.validate_string_args(expected , 'meh')
+        self.assertEqual(expected, ret)
+
+    def test_multi(self):
+        expected = ['abc', 'def']
+        returned = dtype_utils.validate_string_args(['   ' + expected[0], expected[1] + '    '], ['meh', 'foo'])
+        for exp, ret in zip(expected, returned):
+            self.assertEqual(exp, ret)
+
+    def test_not_string_lists(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.validate_string_args([14], ['meh'])
+
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.validate_string_args(14, ['meh'])
+
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.validate_string_args({'dfdf': 14}, ['meh'])
+
+    def test_name_not_string(self):
+        actual = ['ghghg']
+        ret = dtype_utils.validate_string_args(actual, [np.arange(3)])
+        self.assertEqual(ret, actual)
+
+    def test_unequal_lengths(self):
+        expected = ['a', 'b']
+        actual = dtype_utils.validate_string_args(expected + ['c'], ['a', 'b'])
+        for exp, ret in zip(expected, actual):
+            self.assertEqual(exp, ret)
+
+    def test_names_not_list_of_strings(self):
+        with self.assertRaises(TypeError):
+            _ = dtype_utils.validate_string_args(['a', 'v'], {'a': 1, 'v': 43})
+
+
 if __name__ == '__main__':
     unittest.main()
